#include "md5.h"

MD5::MD5()
{
	this->init();
}

MD5::~MD5()
{
}

void MD5::init()
{
	this->bytes = 0;
	this->work[0] = 0x67452301;
	this->work[1] = 0xefcdab89;
	this->work[2] = 0x98badcfe;
	this->work[3] = 0x10325476;
}

void MD5::update(std::string &buf)
{
	this->update((const uint8_t *)buf.c_str(), buf.size());
}

void MD5::update(const  int8_t *buf, int size)
{
	this->update((const uint8_t *)buf, size);
}

void MD5::update(const uint8_t *buf, int size)
{
	int n;
	int offset;

	offset = (int)(this->bytes & 63);
	if(offset){
		n = 64-offset;
		if(size < n){
			n = size;
		}
		if(n){
			memcpy(this->buff+offset, buf, n);
			offset += n;
			size -= n;
			buf += n;
			this->bytes += n;
		}
		if(offset == 64){
			this->process(this->buff);
		}else{
			return;
		}
	}

	while(size >= 64){
		this->process(buf);
		size -= 64;
		buf += 64;
		this->bytes += 64;
	}

	if(size){
		memcpy(this->buff, buf, size);
		this->bytes += size;
	}
}

std::string MD5::digest()
{
	std::string r;
	uint8_t buf[16];

	this->digest(buf);
	r.assign((const char *)buf, 16);
	
	return r;
}

void MD5::digest( int8_t *dst)
{
	this->digest((uint8_t *)dst);
}

void MD5::digest(uint8_t *dst)
{
	int i,j;
	int n;
	
	uint64_t bits;
	uint8_t buf[8];

	static const uint8_t padding[64] = {
		0x80, 0, 0, 0, 0, 0, 0, 0, // 0-7
		0, 0, 0, 0, 0, 0, 0, 0,    // 8-15
		0, 0, 0, 0, 0, 0, 0, 0,    // 16-23
		0, 0, 0, 0, 0, 0, 0, 0,    // 24-31
		0, 0, 0, 0, 0, 0, 0, 0,    // 32-39
		0, 0, 0, 0, 0, 0, 0, 0,    // 40-47
		0, 0, 0, 0, 0, 0, 0, 0,    // 48-55
		0, 0, 0, 0, 0, 0, 0, 0,    // 56-63
	};

	bits = this->bytes << 3;
	for(i=0;i<8;i++){
		buf[i] = (uint8_t)( (bits>>(i*8)) & 0xff );
	}

	n = (int)(this->bytes & 63);
	if(n <= 56){
		this->update(padding, 56-n);
	}else{
		this->update(padding, 120-n);
	}

	this->update(buf, 8);

	for(i=0;i<4;i++){
		for(j=0;j<4;j++){
			dst[i*4+j] = (uint8_t)( (this->work[i]>>(j*8)) & 0xff );
		}
	}
}

static inline uint32_t rol(uint32_t src, int bits)
{
	return ((src<<bits) | (src>>(32-bits)));
}

static inline uint32_t F(uint32_t x, uint32_t y, uint32_t z)
{
	return ((x & y) | ((~x) & z));
}

static inline uint32_t round_1(uint32_t a, uint32_t b, uint32_t c, uint32_t d, uint32_t s, uint32_t t, uint32_t x)
{
	uint32_t w;

	w = a + F(b,c,d) + x + t;

	return (b + rol(w, s));
}

static inline uint32_t G(uint32_t x, uint32_t y, uint32_t z)
{
	return ((x & z) | (y & (~z)));
}

static inline uint32_t round_2(uint32_t a, uint32_t b, uint32_t c, uint32_t d, uint32_t s, uint32_t t, uint32_t x)
{
	uint32_t w;

	w = a + G(b,c,d) + x + t;

	return (b + rol(w, s));
}

static inline uint32_t H(uint32_t x, uint32_t y, uint32_t z)
{
	return (x ^ y ^ z);
}

static inline uint32_t round_3(uint32_t a, uint32_t b, uint32_t c, uint32_t d, uint32_t s, uint32_t t, uint32_t x)
{
	uint32_t w;

	w = a + H(b,c,d) + x + t;

	return (b + rol(w, s));
}

static inline uint32_t I(uint32_t x, uint32_t y, uint32_t z)
{
	return (y ^ (x | (~z)));
}

static inline uint32_t round_4(uint32_t a, uint32_t b, uint32_t c, uint32_t d, uint32_t s, uint32_t t, uint32_t x)
{
	uint32_t w;

	w = a + I(b,c,d) + x + t;

	return (b + rol(w, s));
}

void MD5::process(const uint8_t *buf)
{
	int i;
	
	uint32_t a,b,c,d;
	uint32_t x[16];

	for(i=0;i<16;i++){
		x[i] = buf[0] + (buf[1] << 8) + (buf[2] << 16) + (buf[3] << 24);
		buf += 4;
	}

	a = this->work[0];
	b = this->work[1];
	c = this->work[2];
	d = this->work[3];

	a = round_1(a,b,c,d,  7, 0xd76aa478, x[ 0]);
	d = round_1(d,a,b,c, 12, 0xe8c7b756, x[ 1]);
	c = round_1(c,d,a,b, 17, 0x242070db, x[ 2]);
	b = round_1(b,c,d,a, 22, 0xc1bdceee, x[ 3]);
	a = round_1(a,b,c,d,  7, 0xf57c0faf, x[ 4]);
	d = round_1(d,a,b,c, 12, 0x4787c62a, x[ 5]);
	c = round_1(c,d,a,b, 17, 0xa8304613, x[ 6]);
	b = round_1(b,c,d,a, 22, 0xfd469501, x[ 7]);
	a = round_1(a,b,c,d,  7, 0x698098d8, x[ 8]);
	d = round_1(d,a,b,c, 12, 0x8b44f7af, x[ 9]);
	c = round_1(c,d,a,b, 17, 0xffff5bb1, x[10]);
	b = round_1(b,c,d,a, 22, 0x895cd7be, x[11]);
	a = round_1(a,b,c,d,  7, 0x6b901122, x[12]);
	d = round_1(d,a,b,c, 12, 0xfd987193, x[13]);
	c = round_1(c,d,a,b, 17, 0xa679438e, x[14]);
	b = round_1(b,c,d,a, 22, 0x49b40821, x[15]);

	a = round_2(a,b,c,d,  5, 0xf61e2562, x[ 1]);
	d = round_2(d,a,b,c,  9, 0xc040b340, x[ 6]);
	c = round_2(c,d,a,b, 14, 0x265e5a51, x[11]);
	b = round_2(b,c,d,a, 20, 0xe9b6c7aa, x[ 0]);
	a = round_2(a,b,c,d,  5, 0xd62f105d, x[ 5]);
	d = round_2(d,a,b,c,  9, 0x02441453, x[10]);
	c = round_2(c,d,a,b, 14, 0xd8a1e681, x[15]);
	b = round_2(b,c,d,a, 20, 0xe7d3fbc8, x[ 4]);
	a = round_2(a,b,c,d,  5, 0x21e1cde6, x[ 9]);
	d = round_2(d,a,b,c,  9, 0xc33707d6, x[14]);
	c = round_2(c,d,a,b, 14, 0xf4d50d87, x[ 3]);
	b = round_2(b,c,d,a, 20, 0x455a14ed, x[ 8]);
	a = round_2(a,b,c,d,  5, 0xa9e3e905, x[13]);
	d = round_2(d,a,b,c,  9, 0xfcefa3f8, x[ 2]);
	c = round_2(c,d,a,b, 14, 0x676f02d9, x[ 7]);
	b = round_2(b,c,d,a, 20, 0x8d2a4c8a, x[12]);

	a = round_3(a,b,c,d,  4, 0xfffa3942, x[ 5]);
	d = round_3(d,a,b,c, 11, 0x8771f681, x[ 8]);
	c = round_3(c,d,a,b, 16, 0x6d9d6122, x[11]);
	b = round_3(b,c,d,a, 23, 0xfde5380c, x[14]);
	a = round_3(a,b,c,d,  4, 0xa4beea44, x[ 1]);
	d = round_3(d,a,b,c, 11, 0x4bdecfa9, x[ 4]);
	c = round_3(c,d,a,b, 16, 0xf6bb4b60, x[ 7]);
	b = round_3(b,c,d,a, 23, 0xbebfbc70, x[10]);
	a = round_3(a,b,c,d,  4, 0x289b7ec6, x[13]);
	d = round_3(d,a,b,c, 11, 0xeaa127fa, x[ 0]);
	c = round_3(c,d,a,b, 16, 0xd4ef3085, x[ 3]);
	b = round_3(b,c,d,a, 23, 0x04881d05, x[ 6]);
	a = round_3(a,b,c,d,  4, 0xd9d4d039, x[ 9]);
	d = round_3(d,a,b,c, 11, 0xe6db99e5, x[12]);
	c = round_3(c,d,a,b, 16, 0x1fa27cf8, x[15]);
	b = round_3(b,c,d,a, 23, 0xc4ac5665, x[ 2]);

	a = round_4(a,b,c,d,  6, 0xf4292244, x[ 0]);
	d = round_4(d,a,b,c, 10, 0x432aff97, x[ 7]);
	c = round_4(c,d,a,b, 15, 0xab9423a7, x[14]);
	b = round_4(b,c,d,a, 21, 0xfc93a039, x[ 5]);
	a = round_4(a,b,c,d,  6, 0x655b59c3, x[12]);
	d = round_4(d,a,b,c, 10, 0x8f0ccc92, x[ 3]);
	c = round_4(c,d,a,b, 15, 0xffeff47d, x[10]);
	b = round_4(b,c,d,a, 21, 0x85845dd1, x[ 1]);
	a = round_4(a,b,c,d,  6, 0x6fa87e4f, x[ 8]);
	d = round_4(d,a,b,c, 10, 0xfe2ce6e0, x[15]);
	c = round_4(c,d,a,b, 15, 0xa3014314, x[ 6]);
	b = round_4(b,c,d,a, 21, 0x4e0811a1, x[13]);
	a = round_4(a,b,c,d,  6, 0xf7537e82, x[ 4]);
	d = round_4(d,a,b,c, 10, 0xbd3af235, x[11]);
	c = round_4(c,d,a,b, 15, 0x2ad7d2bb, x[ 2]);
	b = round_4(b,c,d,a, 21, 0xeb86d391, x[ 9]);

	this->work[0] += a;
	this->work[1] += b;
	this->work[2] += c;
	this->work[3] += d;
}
